# PRG - Exam Test #

This is the assignment for the PRG exam test. Follow the instructions precisely and do not add
any additional code to the modules implemented (not even print statements) as they will be imported
and evaluated automatically. Syntactic errors in a module result in 0 points from the assignment,
as all automatic tests will fail on that module.

The modules (packages) you implement should be packed in a zip file (files should be in the root of
the zip file, zipping directly the files, not the containing folder) and uploaded to BRUTE. For
every module, class and function you implement, do not forget to add a docstring. Code quality
(mainly following naming conventions and writing docstrings) is part of the grading of your
solution.


## Task 1 [2p] ##

Consider the following sequence of python statements

1. list_a = [1, 2, 3]
2. list_b = list_a
3. list_b.append(4)
4. list_c = list_a.copy()

Explain what happens on the stack and heap when each of the statements 2-4 gets executed. Put
your explanation in a file "stack.txt".


## Task 2 [6p] ##

Create a module named "task2_functions" and implement three functions according to the following
unit tests. You can put the unit tests (from below) to the same module.

```
import unittest

class TestFunctions(unittest.TestCase):

    def test_list_of_multiplies(self):
        # [2p]
        self.assertEqual([3, 6, 9], list_of_multiplies(3, length=3))
        self.assertEqual([2, 4, 6, 8, 10, 12, 14, 16, 18, 20], list_of_multiplies(2, length=10))
        self.assertEqual([12, 24, 36, 48, 60], list_of_multiplies(12, length=5))
        self.assertEqual([], list_of_multiplies(3, length=0))
        with self.assertRaises(ValueError):
            list_of_multiplies(3, length=-2)

    def test_pluralize_words(self):
        # [2.5p]
        self.assertEqual(["bird", "cats"], pluralize_words(["cat", "bird", "cat"]))
        self.assertEqual(["dogs"], pluralize_words(["dog", "dog", "dog"]))
        self.assertEqual(["frog", "snake"], pluralize_words(["snake", "frog"]))

    def test_string_add(self):
        # [1.5p]
        self.assertEqual("90", string_add("10", "80"))
        self.assertEqual("10", string_add("12", "-2"))
        self.assertEqual("nan", string_add("12", "0-"))
        self.assertEqual("nan", string_add("12", "a2"))
```


## Task 3 [5p] ##

Create a module named "tuple_io", which will contain two functions. These two function will save and
load a tuple of integers in a text file. The format of the file is up to you, only make sure that
the load function can load what the save function has saved.

1. [2p] Implement function save_tuple, which takes two arguments - a tuple of numbers and a file
    path, and will store the tuple under given path.
2. [1p] In function save_tuple, if the folder where the data file should be created, does not exist,
    skip saving the data without raising an exception.
3. [2p] Implement function load_tuple, which takes a single argument, a file path, and returns
    a tuple loaded. If the file does not exist, return an empty tuple.


## Task 4 [5p] ##

Create a package containing a single module. In this module, implement two functions operating on
strings.

1. [1p] Choose the name of the package, module and function appropriately based on the
    functionality defined below.
2. [1p] The first function is named "is_palindrome" and takes a single argument, a word, and returns
    whether the word is a palindrome (e.g. "madam" -> True, "sir" -> False)
3. [1p] The second function takes a single argument, space-separated upper-case words, and returns
    an English sentence where the first letter is upper-case, all other letters are lower-case and
    there is a period at the end (e.g. "THE VIEW IS FANTASTIC" -> "The view is fantastic.").
4. [2p] On the same level as the package (top-level folder), create a module containing unit tests
    for the implemented module. Make sure that the tests pass for your functions.


## Task 5 [7p] ##

Create a module named "library".

1. [1p] Implement class Authors that stores a list of authors (strings) for a book.
2. [1p] Enable class instantiation without a parameter as well.
3. [2p] Implement support for the plus operator, where the first argument is always Authors
    instance and second argument can be either string or also Authors instance. The result of
    the addition is a new Authors instance.
4. [1p] Implement the equality operator - two Authors instances are equal if and only if the
    lists of authors are equal. The order of the authors does not matter and multiplicities do not
    matter - if an author is multiple times in the list, it is treated as if it is there once.
5. [2p] Implement a class SingleNameAuthors that inherits from Authors and where each string
    in the list of authors contains a single word (e.g. family name). Ensure this by raising
    an appropriate exception. Try to reuse as much code from Authors as possible.

Authors example:

```
>>> Authors(["Vladimir Nabokov", "James Joyce"]) == Authors(["James Joyce", "Vladimir Nabokov", "James Joyce"])
True
>>> Authors(["Vladimir Nabokov", "James Joyce"]) == Authors() + "James Joyce" + Authors(["Vladimir Nabokov"])
True
```


## Epilogue ##

At the end of the test, the following files (folders) should be present in your zip file. They
should be in the root of the zip file, not in a nested folder.

- [folder of chosen name] - task 4
- library.py - task 5
- tuple_io.py - task 3
- task2_functions.py - task 2
- test_[module of chosen name].py - task 4
- stack.txt - task 1

